from typing import *
from adsg_core.graph.adsg import *
from adsg_core.graph.sup.nodes import *
from adsg_core.graph.adsg_basic import *
from adsg_core.graph.adsg_nodes import *
from adsg_core.graph.graph_edges import *

__all__ = [
    'SupDSG', 'EdgeType', 'DSGType', 'SupNode',
    'ConnNodes', 'SelectionChoiceNode', 'ConnectionChoiceNode', 'ConnectorNode', 'ConnectorDegreeGroupingNode',
    'SupChoiceMapping', 'SupInitializationError', 'SupResolveError', 'SupSelChoiceOptionMapping',
    'SupExistenceMapping',
]


class SupDSG(BasicDSG):
    """
    Supplementary Design Space Graph: a DSG where all choices are fully derived from another DSG (the source DSG).
    Can be useful for modeling variability in graph structures other than the source DSG.

    Usage:
    1. Add edges and choices (add_edge, add_edges, add_selection_choice, add_connection_choice)
    2. Add choice mappings (add_mapping)
    3. Set start nodes and initialize choices (set_start_nodes)
    """

    def __init__(self, *args, choice_mappings=None, **kwargs):
        self._choice_mappings: List[Tuple[ChoiceNode, SupChoiceMapping]] = choice_mappings or []
        super().__init__(*args, **kwargs)

    def _mod_graph_adjust_kwargs(self, kwargs):
        super()._mod_graph_adjust_kwargs(kwargs)
        kwargs['choice_mappings'] = self._choice_mappings

    def _mod_graph_inplace(self, kwargs):
        if 'choice_mappings' in kwargs:
            self._choice_mappings = kwargs['choice_mappings']

    @property
    def sup_nodes(self) -> List[SupNode]:
        return self.get_nodes_by_type(SupNode)

    def get_by_ref(self, ref) -> Optional[SupNode]:
        for node in self.sup_nodes:
            if node.ref == ref:
                return node

    @property
    def choice_mappings(self) -> List[Tuple[ChoiceNode, 'SupChoiceMapping']]:
        return self._choice_mappings

    def add_mapping(self, sup_choice_node: ChoiceNode, src_dsg: DSG, choice_mapping: 'SupChoiceMapping'):
        """Add a choice mapping to the Supplementary DSG for a given source DSG"""
        if sup_choice_node not in self.graph.nodes:
            raise RuntimeError(f'Choice node not found: {sup_choice_node!r}')

        choice_mapping.initialize(self, sup_choice_node, src_dsg)
        self._choice_mappings.append((sup_choice_node, choice_mapping))

    def initialize_choices(self):
        mapped_choice_nodes = set()
        dup_mapped = []
        for choice_node, _ in self._choice_mappings:
            if choice_node in mapped_choice_nodes:
                dup_mapped.append(choice_node)
            else:
                mapped_choice_nodes.add(choice_node)

        if len(dup_mapped) > 0:
            raise RuntimeError(f'Duplicate mapped choice nodes: {dup_mapped!r}')

        unmapped_choice_nodes = set(self.choice_nodes) - mapped_choice_nodes
        if len(unmapped_choice_nodes):
            raise RuntimeError(f'Unmapped choice nodes: {unmapped_choice_nodes!r}')

        return super().initialize_choices()

    def resolve(self, src_dsg: DSG) -> 'SupDSG':
        """Resolve choice mappings from a finalized source DSG"""
        if not src_dsg.final or not src_dsg.feasible:
            raise RuntimeError('Expecting a final and feasible source DSG')

        sup_dsg = self
        for choice_node, choice_mapping in self.choice_mappings:
            # Resolve if choice not is active in the SupDSG
            if choice_node in sup_dsg.graph.nodes:
                sup_dsg = choice_mapping.resolve(sup_dsg, choice_node, src_dsg)

        if not sup_dsg.final:
            raise RuntimeError('Resolved SupDSG is not final; choice nodes remain!')
        return sup_dsg

    def constrain_choices(self, *args, **kwargs):
        raise RuntimeError('Supplementary DSG does not support choice constraints')

    def add_incompatibility_constraint(self, *args, **kwargs):
        raise RuntimeError('Supplementary DSG does not support incompatibility constraints')


class SupInitializationError(RuntimeError):

    def __init__(self, mapping: 'SupChoiceMapping', sup_dsg: SupDSG, source_dsg: DSG, msg: str):
        self.mapping = mapping
        self.sup_dsg = sup_dsg
        self.source_dsg = source_dsg
        super().__init__(msg)


class SupResolveError(RuntimeError):

    def __init__(self, mapping: 'SupChoiceMapping', sup_dsg: SupDSG, source_dsg: DSG, msg: str):
        self.mapping = mapping
        self.sup_dsg = sup_dsg
        self.source_dsg = source_dsg
        super().__init__(msg)


class SupChoiceMapping:
    """
    Base class for defining a choice mapping in a supplementary design space graph.
    """

    def initialize(self, sup_dsg: SupDSG, sup_choice_node: ChoiceNode, src_dsg: DSG):
        """
        Called after adding the choice mapping to a Supplementary DSG.
        Supplementary DSG and source DSG are in their base states (including choices etc.).
        Should raise a SupInitializationError if it cannot be initialized.
        """

    def resolve(self, sup_dsg: SupDSG, sup_choice_node: ChoiceNode, src_dsg: DSG) -> SupDSG:
        """
        Given the associated choice node in a Supplementary DSG, should resolve the mapping from the source DSG.
        Supplementary DSG and source DSG may be in partially-resolve states already.
        Returns the resolved Supplementary DSG.
        Should raise a SupResolveError if it cannot be resolved.
        """
        raise NotImplementedError

    def __str__(self):
        raise NotImplementedError

    def __repr__(self):
        raise NotImplementedError


class SupSelChoiceOptionMapping(SupChoiceMapping):
    """
    Maps each option of a selection choice of the source DSG to an option of a selection choice of a supplementary DSG.

    `None` as a mapping source value indicates the situation where the choice in the source DSG was not active (i.e. the
    originating node does not exist anymore in the source DSG).
    """

    def __init__(self, src_choice_node: SelectionChoiceNode, mapping: Dict[Optional[DSGNode], DSGNode]):
        self._src_choice_originating_node = None
        self._src_choice_node = src_choice_node
        self._mapping = mapping

    def initialize(self, sup_dsg: SupDSG, sup_choice_node: ChoiceNode, src_dsg: DSG):
        if not isinstance(sup_choice_node, SelectionChoiceNode):
            raise SupInitializationError(self, sup_dsg, src_dsg, f'Expecting SelectionChoiceNode: {sup_choice_node!r}')

        # Check if all source choice option nodes are mapped
        src_choice_node = self._src_choice_node
        if src_choice_node not in src_dsg.graph.nodes:
            raise SupInitializationError(
                self, sup_dsg, src_dsg, f'Source choice node not in source DSG: {src_choice_node!r}')
        src_option_nodes = src_dsg.get_option_nodes(src_choice_node)

        mapping_nodes = {key for key in self._mapping.keys() if key is not None}
        unmapped_src_opt_nodes = set(src_option_nodes) - mapping_nodes
        if len(unmapped_src_opt_nodes) > 0:
            raise SupInitializationError(
                self, sup_dsg, src_dsg, f'Not all source choice option nodes mapped: {unmapped_src_opt_nodes!r}')

        if src_dsg.has_conditional_existence(src_choice_node) and None not in self._mapping:
            raise SupInitializationError(self, sup_dsg, src_dsg,
                                         f'Choice may be inactive; `None` missing from mapping!')

        # Check if all target sup option nodes are available
        sup_option_nodes = sup_dsg.get_option_nodes(sup_choice_node)
        unknown_sup_opt_nodes = set(self._mapping.values()) - set(sup_option_nodes)
        if len(unknown_sup_opt_nodes) > 0:
            raise SupInitializationError(
                self, sup_dsg, src_dsg, f'Not all target nodes are choice option nodes: {unknown_sup_opt_nodes!r}')

        # Get originating node in source DSG
        originating_nodes = [edge[0] for edge in iter_in_edges(src_dsg.graph, src_choice_node)]
        self._src_choice_originating_node = originating_nodes[0]

    def resolve(self, sup_dsg: SupDSG, sup_choice_node: ChoiceNode, src_dsg: DSG) -> SupDSG:
        assert isinstance(sup_choice_node, SelectionChoiceNode)
        src_originating_node = self._src_choice_originating_node
        assert src_originating_node is not None

        # Determine which option has been selected
        src_nodes = {node.str_context() for node in src_dsg.graph.nodes if isinstance(node, DSGNode)}
        mapping = self._mapping
        if src_originating_node.str_context() not in src_nodes:
            if None not in mapping:
                raise SupResolveError(
                    self, sup_dsg, src_dsg, f'Could not resolve {mapping!r}: {self._src_choice_node!r} is inactive, '
                                            f'but `None` is missing from the mapping')
            sup_tgt_option_node = mapping[None]

        else:
            originating_out_nodes = {
                edge[1] for edge in iter_out_edges(src_dsg.graph, src_originating_node, edge_type=EdgeType.DERIVES)}
            src_selected_opt_nodes = originating_out_nodes & set(mapping.keys())
            if len(src_selected_opt_nodes) != 1:
                raise SupResolveError(
                    self, sup_dsg, src_dsg, f'Could not resolve {mapping!r}: could not determine which option node '
                                            f'was selected ({src_selected_opt_nodes!r})')

            mapping_ctx = {node.str_context(): sup_node for node, sup_node in mapping.items()}
            sup_tgt_option_node = mapping_ctx[list(src_selected_opt_nodes)[0].str_context()]

        # Apply selection choice node
        return sup_dsg.get_for_apply_selection_choice(sup_choice_node, sup_tgt_option_node)

    def __str__(self):
        mapping_str = '; '.join([f'{src_node!s} -> {tgt_node!s}' for src_node, tgt_node in self._mapping.items()])
        return f'Sel Choice Opt Mapping {self._src_choice_node!s}: {mapping_str}'

    def __repr__(self):
        return f'{self.__class__.__name__}(src_choice_node={self._src_choice_node!r}, mapping={self._mapping!r})'


class SupExistenceMapping(SupChoiceMapping):
    """
    Maps any node in the source DSG to options of a selection choice of a supplementary DSG. Determines which option to
    select based on source DSG node existence. Since node existence is not mutually-exclusive, the order in which
    mappings are declared determines selection priority:
    - If the first source node is selected, the first option is selected
    - Else if the second node exists, the second is selected
    - Else if ... third

    The option for the case where none of the nodes exist is specified by `None`.
    """

    def __init__(self, mapping: Dict[Optional[DSGNode], DSGNode]):
        self._mapping = mapping

    def initialize(self, sup_dsg: SupDSG, sup_choice_node: ChoiceNode, src_dsg: DSG):
        if not isinstance(sup_choice_node, SelectionChoiceNode):
            raise SupInitializationError(self, sup_dsg, src_dsg, f'Expecting SelectionChoiceNode: {sup_choice_node!r}')

        # Check if all source nodes exist
        mapping_nodes = {key for key in self._mapping.keys() if key is not None}
        unknown_src_nodes = mapping_nodes - set(src_dsg.graph.nodes)
        if len(unknown_src_nodes) > 0:
            raise SupInitializationError(
                self, sup_dsg, src_dsg, f'Source nodes not in source DSG: {unknown_src_nodes!r}')

        # Ensure None is present
        if None not in self._mapping:
            raise SupInitializationError(
                self, sup_dsg, src_dsg, 'Case where all nodes do not exist (`None`) is not specified!')

        # Check if all target sup option nodes are available
        sup_option_nodes = sup_dsg.get_option_nodes(sup_choice_node)
        unknown_sup_opt_nodes = set(self._mapping.values()) - set(sup_option_nodes)
        if len(unknown_sup_opt_nodes) > 0:
            raise SupInitializationError(
                self, sup_dsg, src_dsg, f'Not all target nodes are choice option nodes: {unknown_sup_opt_nodes!r}')

    def resolve(self, sup_dsg: SupDSG, sup_choice_node: ChoiceNode, src_dsg: DSG) -> SupDSG:
        assert isinstance(sup_choice_node, SelectionChoiceNode)

        # Determine selected option node
        src_nodes = {node.str_context() for node in src_dsg.graph.nodes if isinstance(node, DSGNode)}
        sup_tgt_option_node = None
        for src_node, sup_option_node in self._mapping.items():
            if src_node is None:
                continue

            if src_node.str_context() in src_nodes:
                sup_tgt_option_node = sup_option_node
                break

        if sup_tgt_option_node is None:
            sup_tgt_option_node = self._mapping[None]

        # Apply selection choice node
        return sup_dsg.get_for_apply_selection_choice(sup_choice_node, sup_tgt_option_node)

    def __str__(self):
        mapping_str = '; '.join([f'{src_node!s} -> {tgt_node!s}' for src_node, tgt_node in self._mapping.items()])
        return f'Existence Mapping: {mapping_str}'

    def __repr__(self):
        return f'{self.__class__.__name__}(mapping={self._mapping!r})'



"""
MIT License

Copyright: (c) 2024, Deutsches Zentrum fuer Luft- und Raumfahrt e.V.
Contact: jasper.bussemaker@dlr.de

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""
import numpy as np
from typing import *
from adsg_core.graph.adsg_nodes import *
from adsg_core.graph.adsg import DSGType
from adsg_core.graph.choice_constraints import *
from adsg_core.graph.choices import NoOptionError
from adsg_core.optimization.hierarchy.base import *
from adsg_core.optimization.hierarchy.registry import SelChoiceEncoderType

__all__ = ['FastHierarchyAnalyzer']


class FastHierarchyAnalyzer(HierarchyAnalyzerBase):
    """
    Hierarchy analyzer that ignores most of the choice interactions and thereby cannot provide some information,
    however should work quickly for any design space size.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._imputation_cache = {}

    def get_encoder_type(self) -> SelChoiceEncoderType:
        return SelChoiceEncoderType.FAST

    def _get_n_opts(self) -> List[int]:
        sel_choice_opt_nodes = self.selection_choice_option_nodes
        return [len(sel_choice_opt_nodes[node]) for node in self.selection_choice_nodes]

    def _get_selection_choice_is_forced(self) -> np.ndarray:
        is_forced = np.array([False for _ in range(len(self.selection_choice_option_nodes))], dtype=bool)

        # For linked choice constraints, set all but one to forced
        i_choice_nodes = {node: i for i, node in enumerate(self.selection_choice_nodes)}
        for choice_constraint in self.adsg.get_choice_constraints():
            if choice_constraint.type == ChoiceConstraintType.LINKED:
                i_choices = sorted([i_choice_nodes[node] for node in choice_constraint.nodes if node in i_choice_nodes])
                if len(i_choices) > 1:
                    for i_dep in i_choices[1:]:
                        is_forced[i_dep] = True

        return is_forced

    def _get_n_combinations(self) -> int:
        n_opts = list(self.n_opts)

        # Apply choice constraints
        permanent_nodes = self.influence_matrix.permanent_nodes_incl_choice_nodes
        i_choice_nodes = {node: i for i, node in enumerate(self.selection_choice_nodes)}
        for choice_constraint in self.adsg.get_choice_constraints():
            i_choices = sorted([i_choice_nodes[node] for node in choice_constraint.nodes if node in i_choice_nodes])
            if len(i_choices) <= 1:
                continue

            is_all_permanent = all([node in permanent_nodes for node in choice_constraint.nodes])
            n_opts[i_choices[0]] = count_n_combinations_max(choice_constraint, is_all_permanent=is_all_permanent)
            for i_other in i_choices[1:]:
                n_opts[i_other] = 1

        # Cartesian product as upper bound
        return int(np.prod(n_opts, dtype=float))

    def get_graph(self, opt_idx: List[int], mask: np.ndarray = None, is_fixed: List[bool] = None, exclude: set = None) \
            -> Tuple[DSGType, List[int], List[bool], Optional[int]]:

        if exclude is None:
            exclude = set()
        if is_fixed is None:
            is_fixed = [False for _ in range(len(opt_idx))]

        graph_cache = self._graph_cache
        sel_choice_nodes = self.selection_choice_nodes
        i_sel_choice_nodes = {node: i for i, node in enumerate(sel_choice_nodes)}
        sel_choice_opt_nodes = self.selection_choice_option_nodes

        def _get_graph() -> Tuple[Tuple[int, ...], DSGType]:
            graph = self.adsg
            if len(opt_idx_try) == 0:
                return graph.copy()

            # Generate graph
            taken_sel_opt = [X_INACTIVE_VALUE for _ in range(len(opt_idx_try))]
            while True:
                # Get next selection-choice node
                choice_nodes = graph.get_ordered_next_choice_nodes()
                if len(choice_nodes) == 0:
                    break
                choice_node = choice_nodes[0]
                if not isinstance(choice_node, SelectionChoiceNode):
                    break

                # Get assigned option
                i_choice = i_sel_choice_nodes[choice_node]
                i_opt = opt_idx_try[i_choice]
                if i_opt == X_INACTIVE_VALUE:
                    raise RuntimeError(f'Unexpected inactive choice: {i_choice} @ {choice_opt_idx}')
                taken_sel_opt[i_choice] = i_opt

                # Check if the graph has already been created
                cache_key = tuple(taken_sel_opt)
                if cache_key in graph_cache:
                    graph = graph_cache[cache_key]
                    continue

                # Make choice
                option_node = sel_choice_opt_nodes[choice_node][i_opt]
                graph_cache[cache_key] = graph = graph.get_for_apply_selection_choice(choice_node, option_node)

            # Verify that indeed no selection_choices are left
            if len([node for node in graph.choice_nodes if isinstance(node, SelectionChoiceNode)]) > 0:
                raise RuntimeError(f'Selection-choice nodes left for dv: {opt_idx}')
            return tuple(taken_sel_opt), graph

        # Iterate over current and neighboring design vectors
        tried = []
        opt_idx_imp = opt_idx
        graph_instance = None
        for opt_idx_try in self._iter_neighborhood(opt_idx, is_fixed):
            tried.append(opt_idx_try)
            if opt_idx_try in exclude:
                continue

            # Check cache
            if opt_idx_try in self._imputation_cache:
                outputs = self._imputation_cache[opt_idx_try]
                if tuple(outputs[1]) not in exclude:
                    return outputs

            # Try to get graph
            try:
                opt_idx_imp, graph_instance = _get_graph()

            except NoOptionError:  # Can happen if due to some choice constraint options are removed
                exclude.add(opt_idx_try)
                continue

            tried.append(opt_idx_imp)
            if opt_idx_imp in exclude:
                exclude.add(opt_idx_try)
                continue

            if graph_instance.feasible:
                break

            # Mark as infeasible
            exclude.add(opt_idx_try)
            exclude.add(opt_idx_imp)

        if not graph_instance.feasible:
            raise RuntimeError('No more feasible graphs!')

        # Update imputation cache
        choice_opt_idx = list(opt_idx_imp)
        activeness = list(np.array(choice_opt_idx) != X_INACTIVE_VALUE)
        outputs = (graph_instance, choice_opt_idx, activeness, None)
        for key in tried:
            self._imputation_cache[key] = outputs
        return outputs

    def _iter_neighborhood(self, opt_idx: List[int], is_fixed: List[bool]) -> Generator[Tuple[int, ...], None, None]:
        n_opts = self.n_opts

        def _iter_values(i_dv):
            i_current = opt_idx[i_dv]
            yield i_current
            if is_fixed[i_dv]:
                return

            for dist in range(1, n_opts[i_dv]):
                yielded = False
                pos_dir = i_current+dist
                if pos_dir < n_opts[i_dv]:
                    yield pos_dir
                    yielded = True

                neg_dir = i_current-dist
                if neg_dir >= 0:
                    yield neg_dir
                    yielded = True

                if not yielded:
                    break

        def _iter_neighbor_next(i_iter=0, prev_values=None):
            if prev_values is None:
                prev_values = []
            for value in _iter_values(i_dv=i_iter):
                next_values = prev_values+[value]
                if i_iter < len(opt_idx)-1:
                    yield from _iter_neighbor_next(i_iter=i_iter+1, prev_values=next_values)
                else:
                    yield tuple(next_values)

        yield from _iter_neighbor_next()

    def get_opt_idx(self, opt_idx: List[int], mask: np.ndarray = None, is_fixed: List[bool] = None,
                    exclude: set = None) -> Tuple[List[int], List[bool], Optional[int]]:
        _, choice_opt_idx, activeness, i_comb = self.get_graph(opt_idx, mask=mask, is_fixed=is_fixed)
        return choice_opt_idx, activeness, i_comb

    def _get_comb_idx(self, opt_idx: List[int], include_mask: np.ndarray = None) \
            -> Tuple[Optional[int], Optional[np.ndarray]]:
        raise RuntimeError

    def _get_nodes_existence(self, nodes: List[DSGNode], i_comb: int = None) -> Optional[np.ndarray]:
        pass

    def _get_available_combinations_mask(self, fixed_comb_idx: Dict[int, int]) -> Optional[np.ndarray]:
        pass

    def _get_choice_option_indices(self) -> Optional[np.ndarray]:
        pass

    def _get_existence_array(self) -> Optional[np.ndarray]:
        pass



"""
MIT License

Copyright: (c) 2024, Deutsches Zentrum fuer Luft- und Raumfahrt e.V.
Contact: jasper.bussemaker@dlr.de

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""
from typing import *
from adsg_core.graph.adsg import *
from adsg_core.graph.traversal import *
from adsg_core.graph.adsg_nodes import *
from adsg_core.graph.graph_edges import *

__all__ = ['BasicDSG', 'EdgeType', 'CDVNode', 'ChoiceConstraint', 'ChoiceConstraintType', 'DSGType', 'ConnNodes',
           'BasicADSG']

ConnNodes = List[Union[ConnectorNode, Tuple[ConnectorDegreeGroupingNode, List[ConnectorNode]]]]


class BasicDSG(DSG):
    """
    Implementation of the DSG with some helper functions:

    - add_edge and add_edges for adding nodes and edges to the graph
    - add_selection_choice for adding edges representing a selection choice
    - add_connection_choice for adding edges representing a connection choice
    - set_start_nodes for setting the nodes where derivation starts

    It is recommended to use these helper functions to build and initialize the graph:

    1. Add edges and choices (add_edge, add_edges, add_selection_choice, add_connection_choice)
    2. Set start nodes (set_start_nodes)
    3. Initialize choices (initialize_choices)
    """

    def __init__(self, *args, start_nodes=None, **kwargs):
        self._start_nodes = start_nodes
        super().__init__(*args, **kwargs)

    def _mod_graph_adjust_kwargs(self, kwargs):
        kwargs['start_nodes'] = self._start_nodes

    def _mod_graph_inplace(self, kwargs):
        if 'start_nodes' in kwargs:
            self._start_nodes = kwargs['start_nodes']

    def _get_derivation_start_nodes(self) -> Optional[Set[DSGNode]]:
        return self._start_nodes

    def set_start_nodes(self, start_nodes: Set[DSGNode] = None, initialize_choices=True):
        """
        Defines which nodes should be used for starting the existence derivations. If not explicitly defined, uses all
        nodes with no incoming edges.

        Nodes that cannot be derived from any of the starting nodes are removed from the graph.
        """
        if start_nodes is None:
            start_nodes = self._get_alternative_start_nodes()
        if len(start_nodes) == 0:
            raise ValueError('Provide at least one starting node!')
        self._start_nodes = start_nodes

        graph = self.graph

        # Ensure all start nodes are actually in the graph
        missing_start_nodes = {start_node for start_node in start_nodes if start_node not in graph.nodes}
        if len(missing_start_nodes) > 0:
            raise ValueError(f'Nodes not in graph cannot be set as start nodes: {missing_start_nodes}')

        # Remove nodes that are not defined by any of the start nodes
        dsg = self
        removed_edges, removed_nodes = set(), set()
        for floating_node in self._get_floating_nodes():
            if floating_node in start_nodes:
                continue
            removed_nodes.add(floating_node)

            derived_edges, derived_nodes = get_derived_edges_for_node(
                graph, floating_node, start_nodes, removed_edges=removed_edges, removed_nodes=removed_nodes)
            removed_edges |= derived_edges
            removed_nodes |= derived_nodes

        if len(removed_edges) > 0 or len(removed_nodes) > 0:
            dsg = dsg.get_for_adjusted(removed_edges=removed_edges, removed_nodes=removed_nodes)

        if initialize_choices:
            return dsg.initialize_choices()
        return dsg

    def _get_alternative_start_nodes(self) -> Set[DSGNode]:
        return self._get_floating_nodes()

    def _get_floating_nodes(self) -> Set[DSGNode]:
        floating_nodes = set()
        for node in self._graph.nodes:
            for edge in iter_in_edges(self._graph, node):
                if get_edge_type(edge) in {EdgeType.DERIVES, EdgeType.CONNECTS}:
                    break
            else:
                floating_nodes.add(node)
        return floating_nodes

    def _choice_sort_key(self, choice_node: CDVNode) -> tuple:
        type_idx = {SelectionChoiceNode: 0, ConnectionChoiceNode: 1, DesignVariableNode: 2}.get(type(choice_node), 3)
        return type_idx, choice_node.decision_id, getattr(choice_node, 'decision_sort_key', '')

    def add_edge(self, src: DSGNode, tgt: DSGNode, edge_type: EdgeType = EdgeType.DERIVES):
        """Add a directed edge between some source and target nodes"""
        self.add_edges([(src, tgt)], edge_type=edge_type)

    def add_edges(self, edges: List[Tuple[DSGNode, DSGNode]], edge_type: EdgeType = EdgeType.DERIVES):
        """Add multiple edges at a time"""
        for src, tgt in edges:
            add_edge(self._graph, src, tgt, edge_type=edge_type)

    def add_node(self, node: DSGNode):
        """Add a single node to the graph.
        Note that it will be removed if it is not set as a start node, or connected to it (through a derivation edge)
        at some point"""
        self._graph.add_node(node)

    def add_selection_choice(self, choice_id: str, originating_node: DSGNode, option_nodes: List[DSGNode]) \
            -> SelectionChoiceNode:
        """
        A selection choice is a choice between one or more mutually-exclusive option nodes. When choosing one of the
        options, the choice node is removed, and the originating node is connected to the chosen option node.
        """
        choice_node = SelectionChoiceNode(choice_id)

        edges = [(originating_node, choice_node)]
        for i, option_node in enumerate(option_nodes):
            if option_node.option_id is None:
                option_node.option_id = i

            edges.append((choice_node, option_node))

        self.add_edges(edges)

        # Reset influence matrix
        self._influence_matrix = None

        return choice_node

    def add_connection_choice(self, choice_id: str, src_nodes: ConnNodes, tgt_nodes: ConnNodes,
                              exclude: List[Tuple[ConnectorNode, ConnectorNode]] = None) -> ConnectionChoiceNode:
        """
        A connection choice is a choice on how to connect one or more source nodes to one or more target nodes.
        Each connector node can place constraints on the accepted connection degree: one or more specific numbers,
        a range, or a lower limit. Additionally, repeated connections between the same nodes can be allowed or not.

        To model the non-influence of connection order, connector nodes can be grouped by a ConnectorDegreeGroupingNode,
        which as connection degree will have the sum of the connection degrees from underlying nodes, depending on
        whether they exist in an architecture or not.
        """
        graph = self._graph
        choice_node = ConnectionChoiceNode(choice_id)
        grouping_nodes: List[ConnectorDegreeGroupingNode] = []

        def _get_choice_nodes(conn_nodes: ConnNodes):
            if len(conn_nodes) == 0:
                raise ValueError('At least one source and target connector should be fined for connection choice!')

            choice_conn_nodes = []
            for conn_node in conn_nodes:
                if isinstance(conn_node, tuple):
                    if len(conn_node) != 2:
                        raise ValueError(f'Expecting a length-2 tuple: {conn_node!r}')

                    # Grouping nodes are represented by DERIVES edges from the underlying nodes to the grouping node
                    conn_grouping_node, underlying_conn_nodes = conn_node

                    if not isinstance(conn_grouping_node, ConnectorDegreeGroupingNode):
                        raise ValueError(f'Expecting a connector grouping node: {conn_grouping_node!r}')
                    choice_conn_nodes.append(conn_grouping_node)
                    grouping_nodes.append(conn_grouping_node)

                    for underlying_conn_node in underlying_conn_nodes:
                        if not isinstance(underlying_conn_node, ConnectorNode):
                            raise ValueError(f'Expecting a connector node: {underlying_conn_node!r}')

                        add_edge(graph, underlying_conn_node, conn_grouping_node, edge_type=EdgeType.DERIVES)

                else:
                    if not isinstance(conn_node, ConnectorNode):
                        raise ValueError(f'Expecting a connector node: {conn_node!r}')
                    choice_conn_nodes.append(conn_node)

            return choice_conn_nodes

        # Add CONNECTS edges from source to choice to target nodes
        for src_node in _get_choice_nodes(src_nodes):
            add_edge(graph, src_node, choice_node, edge_type=EdgeType.CONNECTS)
        for tgt_node in _get_choice_nodes(tgt_nodes):
            add_edge(graph, choice_node, tgt_node, edge_type=EdgeType.CONNECTS)

        # Add exclusion edges
        if exclude is not None:
            for src, tgt in exclude:
                add_edge(graph, src, tgt, edge_type=EdgeType.EXCLUDES)

        # Update connection degrees for grouping nodes
        for grouping_node in grouping_nodes:
            grouping_node.update_deg(graph)

        return choice_node

    def next(self, node, edge_type: EdgeType = None):
        """Iterate over outgoing nodes, optionally filtering by edge type"""
        for edge in iter_out_edges(self._graph, node, edge_type=edge_type):
            yield edge[1]

    def derived_nodes(self, node):
        """Get all nodes derived by this node"""
        cache = {}
        _, nodes = get_derived_edges_for_node(self._graph, node, start_nodes=None, cache=cache)
        return nodes

    def prev(self, node, edge_type: EdgeType = None):
        """Iterate over incoming nodes, optionally filtering by edge type"""
        for edge in iter_in_edges(self._graph, node, edge_type=edge_type):
            yield edge[0]


BasicADSG = BasicDSG  # Backward compatibility



import itertools
import numpy as np
from adsg_core.optimization.assign_enc.encoding import *
from adsg_core.optimization.assign_enc.eager.encodings.grouped_base import *

__all__ = ['AmountGrouper', 'LocationGrouper', 'AmountFirstGroupedEncoder', 'TotalAmountGrouper', 'SourceAmountGrouper',
           'SourceAmountFlattenedGrouper', 'TargetAmountGrouper', 'TargetAmountFlattenedGrouper',
           'OneVarLocationGrouper', 'FlatIndexLocationGrouper', 'RelFlatIndexLocationGrouper',
           'CoordIndexLocationGrouper', 'RelCoordIndexLocationGrouper']


class AmountGrouper:
    """Base class for grouping by total connection amount."""

    def get_grouping_values(self, matrix: np.ndarray) -> np.ndarray:
        raise NotImplementedError

    def __repr__(self):
        raise NotImplementedError

    def __str__(self):
        raise NotImplementedError


class LocationGrouper:
    """Base class for grouping by connection location."""

    def get_grouping_values(self, matrix: np.ndarray) -> np.ndarray:
        raise NotImplementedError

    def __repr__(self):
        raise NotImplementedError

    def __str__(self):
        raise NotImplementedError


class AmountFirstGroupedEncoder(GroupedEncoder):
    """Grouped encoder that first groups by the total amount of connections to make, and then by the locations of these
    connections."""

    def __init__(self, imputer, amount_grouper: AmountGrouper, loc_grouper: LocationGrouper, **kwargs):
        self.amount_grouper = amount_grouper
        self.loc_grouper = loc_grouper
        super().__init__(imputer, **kwargs)

    def _get_grouping_values(self, matrix: np.ndarray) -> np.ndarray:
        return np.column_stack([
            self.amount_grouper.get_grouping_values(matrix),
            self.loc_grouper.get_grouping_values(matrix),
        ])

    def __repr__(self):
        return f'{self.__class__.__name__}({self._imputer!r}, {self.amount_grouper!r}, {self.loc_grouper!r})'

    def __str__(self):
        return f'Amount First + {self._imputer!s}: {self.amount_grouper!s}; {self.loc_grouper!s}'


class TotalAmountGrouper(AmountGrouper):
    """Group by total amount of connections"""

    def get_grouping_values(self, matrix: np.ndarray) -> np.ndarray:
        return np.sum(flatten_matrix(matrix), axis=1)

    def __repr__(self):
        return f'{self.__class__.__name__}()'

    def __str__(self):
        return f'Total Amount'


class SourceAmountGrouper(AmountGrouper):
    """Group by number of connections from source nodes"""

    def get_grouping_values(self, matrix: np.ndarray) -> np.ndarray:
        return np.sum(matrix, axis=2)

    def __repr__(self):
        return f'{self.__class__.__name__}()'

    def __str__(self):
        return f'Source Amount'


class SourceAmountFlattenedGrouper(AmountGrouper):
    """Group by number of connections from source nodes; summarized in one design variable"""

    def get_grouping_values(self, matrix: np.ndarray) -> np.ndarray:
        n_src = np.sum(matrix, axis=2)
        _, unique_indices = np.unique(n_src, return_inverse=True, axis=0)
        return np.array([unique_indices]).T

    def __repr__(self):
        return f'{self.__class__.__name__}()'

    def __str__(self):
        return f'Source Amount Flat'


class TargetAmountGrouper(AmountGrouper):
    """Group by number of connections to target nodes"""

    def get_grouping_values(self, matrix: np.ndarray) -> np.ndarray:
        return np.sum(matrix, axis=1)

    def __repr__(self):
        return f'{self.__class__.__name__}()'

    def __str__(self):
        return f'Target Amount'


class TargetAmountFlattenedGrouper(AmountGrouper):
    """Group by number of connections to target nodes; summarized in one design variable"""

    def get_grouping_values(self, matrix: np.ndarray) -> np.ndarray:
        n_src = np.sum(matrix, axis=1)
        _, unique_indices = np.unique(n_src, return_inverse=True, axis=0)
        return np.array([unique_indices]).T

    def __repr__(self):
        return f'{self.__class__.__name__}()'

    def __str__(self):
        return f'Target Amount Flat'


class OneVarLocationGrouper(LocationGrouper):
    """One design variable for each remaining matrix after selecting the amounts"""

    def get_grouping_values(self, matrix: np.ndarray) -> np.ndarray:
        return np.arange(matrix.shape[0])

    def __repr__(self):
        return f'{self.__class__.__name__}()'

    def __str__(self):
        return f'One Var Amount'


class FlatIndexLocationGrouper(LocationGrouper):
    """Group by connection location indices. Inefficient for very large nr of matrices."""

    def get_grouping_values(self, matrix: np.ndarray) -> np.ndarray:
        matrix_flat = flatten_matrix(matrix)
        if matrix_flat.shape[0] == 0:
            return np.zeros((0, 0), dtype=int)
        n_conn_max = np.max(np.sum(matrix_flat, axis=1))

        n_mat = matrix.shape[0]
        dv_loc_idx = np.zeros((n_mat, n_conn_max), dtype=int)
        for i_mat in range(n_mat):
            loc_idx = self._get_loc_indices(matrix_flat[i_mat, :])
            dv_loc_idx[i_mat, :len(loc_idx)] = loc_idx
        return dv_loc_idx

    @classmethod
    def _get_loc_indices(cls, conn_arr: np.ndarray) -> np.ndarray:
        conn_arr = conn_arr.copy()
        loc_idx = []
        idx = 0
        while idx < len(conn_arr):
            if conn_arr[idx] > 0:
                loc_idx.append(idx)
                conn_arr[idx] -= 1
            else:
                idx += 1
        return np.array(loc_idx)

    def __repr__(self):
        return f'{self.__class__.__name__}()'

    def __str__(self):
        return f'Flat Idx Loc'


class RelFlatIndexLocationGrouper(FlatIndexLocationGrouper):
    """Group by relative connection location indices. Inefficient for very large nr of matrices."""

    @classmethod
    def _get_loc_indices(cls, conn_arr: np.ndarray) -> np.ndarray:
        abs_loc_idx = super()._get_loc_indices(conn_arr)
        return np.diff(np.concatenate([[0], abs_loc_idx]))

    def __repr__(self):
        return f'{self.__class__.__name__}()'

    def __str__(self):
        return f'Rel Flat Idx Loc'


class CoordIndexLocationGrouper(LocationGrouper):
    """Group by connection location indices (encoded as i_src, i_tgt). Inefficient for very large nr of matrices."""

    def get_grouping_values(self, matrix: np.ndarray) -> np.ndarray:
        matrix_flat = flatten_matrix(matrix)
        if matrix_flat.shape[0] == 0:
            return np.zeros((0, 0), dtype=int)
        n_conn_max = np.max(np.sum(matrix_flat, axis=1))*2

        n_mat, n_src, n_tgt = matrix.shape[0], matrix.shape[1], matrix.shape[2]
        dv_loc_idx = np.zeros((n_mat, n_conn_max), dtype=int)
        idx_map = list(itertools.product(range(n_src), range(n_tgt)))
        for i_mat in range(n_mat):
            loc_idx = self._get_loc_indices(matrix_flat[i_mat, :], idx_map)
            dv_loc_idx[i_mat, :len(loc_idx)] = loc_idx

        dv_loc_idx -= np.min(dv_loc_idx, axis=0)
        return dv_loc_idx

    @classmethod
    def _get_loc_indices(cls, conn_arr: np.ndarray, idx_map) -> np.ndarray:
        conn_arr = conn_arr.copy()
        loc_idx = []
        idx = 0
        while idx < len(conn_arr):
            if conn_arr[idx] > 0:
                loc_idx.append(idx_map[idx])
                conn_arr[idx] -= 1
            else:
                idx += 1
        return np.array([idx for (i_src, i_tgt) in loc_idx for idx in [i_src, i_tgt]]).astype(int)

    def __repr__(self):
        return f'{self.__class__.__name__}()'

    def __str__(self):
        return f'Coord Idx Loc'


class RelCoordIndexLocationGrouper(CoordIndexLocationGrouper):
    """Group by relative connection location indices (encoded as i_src, i_tgt). Inefficient for very large nr of mat."""

    @classmethod
    def _get_loc_indices(cls, conn_arr: np.ndarray, idx_map) -> np.ndarray:
        abs_loc_idx = super()._get_loc_indices(conn_arr, idx_map)
        rel_loc_idx = np.zeros(abs_loc_idx.shape)
        rel_loc_idx[0::2] = np.diff(np.concatenate([[0], abs_loc_idx[0::2]]))
        rel_loc_idx[1::2] = np.diff(np.concatenate([[0], abs_loc_idx[1::2]]))
        return rel_loc_idx

    def __repr__(self):
        return f'{self.__class__.__name__}()'

    def __str__(self):
        return f'Rel Coord Idx Loc'



"""
MIT License

Copyright: (c) 2024, Deutsches Zentrum fuer Luft- und Raumfahrt e.V.
Contact: jasper.bussemaker@dlr.de

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""
import logging
import warnings
import functools
import numpy as np
from typing import *
from concurrent.futures import wait, ProcessPoolExecutor, ThreadPoolExecutor
from adsg_core.optimization.evaluator import DSGEvaluator
from adsg_core.optimization.dv_output_defs import DesVar
from adsg_core.optimization.graph_processor import GraphProcessor
from adsg_core.optimization.assign_enc.time_limiter import run_timeout

try:
    from sb_arch_opt.problem import ArchOptProblemBase
    from sb_arch_opt.design_space import ArchDesignSpace
    from pymoo.core.variable import Variable, Real, Integer, Choice

    from sb_arch_opt.sampling import TrailRepairWarning
    warnings.simplefilter("ignore", category=TrailRepairWarning)

    HAS_SB_ARCH_OPT = True

except ImportError:
    HAS_SB_ARCH_OPT = False

    class ArchDesignSpace:
        pass

    class ArchOptProblemBase:
        pass

__all__ = ['check_dependency', 'DSGDesignSpace', 'DSGArchOptProblem', 'HAS_SB_ARCH_OPT',
           'ADSGDesignSpace', 'ADSGArchOptProblem']

log = logging.getLogger('adsg.opt')


def check_dependency():
    if not HAS_SB_ARCH_OPT:
        raise ImportError('Looks like SBArchOpt is not installed! Run: pip install sb-arch-opt')


class DSGDesignSpace(ArchDesignSpace):
    """
    SBArchOpt design space implementation for a DSG design problem.
    """
    x_all_cutoff = 500000
    x_all_timeout = 10  # sec

    def __init__(self, processor: GraphProcessor):
        if len(processor.des_vars) == 0:
            raise ValueError('No design variables in optimization problem!')

        self._processor = processor
        super().__init__()

    @property
    def processor(self) -> GraphProcessor:
        return self._processor

    def is_explicit(self) -> bool:
        return False

    def _get_variables(self) -> List['Variable']:
        """Returns the list of design variables (pymoo classes)"""
        des_vars = []
        for i, des_var in enumerate(self._processor.des_vars):
            if des_var.is_discrete:
                # Define as categorical is explicitly said so and if there are more than two options (gives slightly
                # better optimizer performance in some cases)
                if self._is_categorical(i, des_var) and len(des_var.options) > 2:
                    des_vars.append(Choice(options=[ii for ii in range(len(des_var.options))]))
                else:
                    des_vars.append(Integer(bounds=(0, len(des_var.options)-1)))

            else:
                des_vars.append(Real(bounds=tuple(des_var.bounds)))

        return des_vars

    def _is_categorical(self, idx, des_var: DesVar) -> bool:
        """Whether a given discrete design variable should be represented as a categorical variable"""
        # We do not have a better way to determine this currently
        return True

    def _is_conditionally_active(self) -> Optional[List[bool]]:
        """Returns for each design variable whether it is conditionally active (i.e. may become inactive)"""
        return self._processor.dv_is_conditionally_active

    def _correct_x(self, x: np.ndarray, is_active: np.ndarray):
        """
        Fill the activeness matrix (n x nx) and if needed correct design vectors (n x nx) that are partially inactive.
        Imputation of inactive variables is handled automatically.
        """
        is_discrete_mask = self.is_discrete_mask
        for i, xi in enumerate(x):
            x_arch = [int(val) if is_discrete_mask[j] else float(val) for j, val in enumerate(xi)]
            _, x_imputed, is_active_arch = self._processor.get_graph(x_arch, create=False)
            x[i, :] = x_imputed
            is_active[i, :] = is_active_arch

    def _quick_sample_discrete_x(self, n: int) -> Tuple[np.ndarray, np.ndarray]:
        """Sample n discrete design vectors (also return is_active) without generating all design vectors first"""
        return self._quick_random_sample_discrete_x(n)

    @functools.lru_cache()
    def _get_n_valid_discrete(self) -> Optional[int]:
        """
        Return the number of valid discrete design points (ignoring continuous dimensions); enables calculation of
        the imputation ratio.
        Valid discrete design points are discrete design points where value constraints are satisfied and where
        inactive design variables are imputed/canonical (compare _get_n_correct_discrete).
        """
        return self._processor.get_n_valid_designs(with_fixed=True)

    @functools.lru_cache()
    def _get_n_active_cont_mean(self) -> Optional[float]:
        """
        Get the mean number of active continuous dimensions, as seen over all valid discrete design vectors.
        """
        _, _, _, _, n_dim_cont_mean, _ = \
            self._processor.get_additional_dv_stats(with_fixed=True)
        return n_dim_cont_mean

    def _get_n_correct_discrete(self) -> Optional[int]:
        """
        Return the number of correct discrete design points (ignoring continuous dimensions); enables calculation of
        the correction ratio.
        Correct discrete design points are discrete design points where value constraints are satisfied, where however
        inactive design variables can have any value (compare _get_n_valid_discrete).
        """

    def _get_n_active_cont_mean_correct(self) -> Optional[float]:
        """
        Get the mean number of active continuous dimensions, as seen over all valid discrete design vectors.
        """

    def _gen_all_discrete_x(self) -> Optional[Tuple[np.ndarray, np.ndarray]]:
        """Generate all possible discrete design vectors (if available). Returns design vectors and activeness
        information."""

        if self._get_n_valid_discrete() > self.x_all_cutoff:
            log.debug(f'Not generating all design vectors because '
                      f'{self._get_n_valid_discrete()} (n_x_discrete) > {self.x_all_cutoff}')
            return

        try:
            log.debug('Generating all design vectors')
            res = run_timeout(
                self.x_all_timeout,
                lambda: self._processor.get_all_discrete_x(with_fixed=True))
            if res is None:
                return
            x, is_active = res
            log.debug(f'Generated {x.shape[0]} design vectors')

            is_cond_act = np.array(self._is_conditionally_active(), dtype=bool)
            any_inactive, not_cond_active = np.any(~is_active, axis=0), ~is_cond_act
            if np.any(any_inactive & not_cond_active):
                i_where, = np.where(any_inactive & not_cond_active)
                raise RuntimeError(f'Inconsistent conditionality vector for i_dv: {i_where}')

            return x, is_active
        except (MemoryError, TimeoutError) as e:
            log.debug(f'Could not generate all design vectors: {e.__class__.__name__}')


class DSGArchOptProblem(ArchOptProblemBase):
    """
    [SBArchOpt](https://sbarchopt.readthedocs.io/) wrapper for a DSG optimization problem. Note that under the
    hood, SBArchOpt uses [pymoo](https://pymoo.org/).
    The connection is made between the `ArchOptProblemBase` class (which specifies all information needed to optimize an
    architecture optimization problem), and the `DSGEvaluator` class, which contains all information for
    running a DSG architecture optimization problem.

    Parallel processing is possible by setting `n_parallel` to a number higher than 1.
    By default, assumes parallel processing is done within the thread and therefore starts a multiprocessing pool to
    run the parallel evaluations.

    Ensure SBArchOpt is installed: `pip install sb-arch-opt`

    Example usage:

    ```python
    from pymoo.optimize import minimize
    from sb_arch_opt.algo.pymoo_interface import get_nsga2

    evaluator = ...  # Instance of DSGEvaluator

    algorithm = get_nsga2(pop_size=100)
    problem = DSGArchOptProblem(evaluator)

    result = minimize(problem, algorithm, termination=('n_eval', 500))
    ```
    """

    def __init__(self, evaluator: DSGEvaluator, n_parallel=None, parallel_processes=True):
        check_dependency()

        self.evaluator = evaluator
        self.n_parallel = n_parallel
        self.parallel_processes = parallel_processes

        n_objs = len(evaluator.objectives)
        n_constr = len(evaluator.constraints)

        design_space = DSGDesignSpace(evaluator)
        super().__init__(design_space, n_obj=n_objs, n_ieq_constr=n_constr)

        self.obj_is_max = [obj.dir.value > 0 for obj in evaluator.objectives]
        self.con_ref = [(con.dir > 0, con.ref) for con in evaluator.constraints]

    def _arch_evaluate(self, x: np.ndarray, is_active_out: np.ndarray, f_out: np.ndarray, g_out: np.ndarray,
                       h_out: np.ndarray, *args, **kwargs):
        # Correct integer design variables
        self.design_space.round_x_discrete(x)

        # Generate architectures
        is_discrete_mask = self.is_discrete_mask
        dsg_instances = []
        for i, xi in enumerate(x):
            x_arch = [int(val) if is_discrete_mask[j] else float(val) for j, val in enumerate(xi)]
            dsg_instance, x_imputed, is_active_arch = self.evaluator.get_graph(x_arch)
            dsg_instances.append(dsg_instance)
            x[i, :] = x_imputed
            is_active_out[i, :] = is_active_arch

        # Evaluate architectures
        if self.n_parallel is not None and self.n_parallel > 1:
            executor_class = ProcessPoolExecutor if self.parallel_processes else ThreadPoolExecutor
            with executor_class(max_workers=self.n_parallel) as executor:
                futures = [executor.submit(self.evaluator.evaluate, dsg) for dsg in dsg_instances]

                wait(futures)
                results = [fut.result() for fut in futures]

        else:
            results = [self.evaluator.evaluate(dsg) for dsg in dsg_instances]

        # Process results
        for i, (obj_values, con_values) in enumerate(results):
            # Correct directions of objectives to represent minimization
            f_out[i, :] = [-val if self.obj_is_max[j] else val for j, val in enumerate(obj_values)]

            # Correct directions and offset constraints to represent g(x) <= 0
            g_out[i, :] = [(val-self.con_ref[j][1])*(-1 if self.con_ref[j][0] else 1)
                           for j, val in enumerate(con_values)]

    def _print_extra_stats(self):
        self.get_discrete_rates(show=True)
        self.evaluator.print_stats()

    def get_n_batch_evaluate(self) -> Optional[int]:
        return self.n_parallel

    def __repr__(self):
        return f'{self.__class__.__name__}({self.evaluator!r})'


ADSGDesignSpace = DSGDesignSpace  # Backward compatibility
ADSGArchOptProblem = DSGArchOptProblem  # Backward compatibility



"""
MIT License

Copyright: (c) 2024, Deutsches Zentrum fuer Luft- und Raumfahrt e.V.
Contact: jasper.bussemaker@dlr.de

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""
import enum
import itertools
import numpy as np
from typing import *
from dataclasses import dataclass
from adsg_core.graph.adsg_nodes import *
from adsg_core.graph.incompatibility import *
from adsg_core.graph.choice_constraints import *
from adsg_core.graph.traversal import traverse_until_choice_nodes
from cached_property import cached_property
if False:
    from adsg_core.graph.adsg import DSGType

__all__ = ['InfluenceMatrix', 'Diag', 'OffDiag', 'StatusScenarios', 'SelectionChoiceScenarios', 'X_INACTIVE_VALUE',
           'X_INACTIVE_IMPUTE']

X_INACTIVE_VALUE = -1
X_INACTIVE_IMPUTE = 0


class Diag(enum.Enum):
    INITIAL = 0  # Non-confirmed initially
    CONFIRMED = 1  # Confirmed
    REMOVED = 2  # Removed
    INFEASIBLE_OPTION = 3  # The option would lead to an infeasible architecture
    CHOICE_MADE = 4  # The decision has been taken (i.e. is not in the graph anymore)


class OffDiag(enum.Enum):
    NO_INTERACTION = 0
    CONFIRMATION = 1  # The option-node at i confirms the existence of the node at j if chosen
    REMOVAL = 2  # The option-node at i removes the node at j if chosen


@dataclass
class StatusScenarios:
    """
    Dataclass representing node status scenarios.
    A status scenario represents all possible combinations of node statuses (e.g. non-confirmed, confirmed, removed)
    given all possible input statuses. Input statuses are determined by non-0 coupling values in the influence matrix.
    """

    # SCENARIO INFO
    nodes: List[DSGNode]  # Nodes this status scenario applies to
    nodes_idx: np.ndarray  # [n_nodes; int] Index of the nodes in the influence matrix

    # INPUT INFLUENCES
    n_scenarios: int  # Nr of input scenarios
    influence_input_node_idx: np.ndarray  # [n_inputs; int] Input node indices (in the influence matrix)
    # Input status matrices, True represents that the input node is selected in the graph, so its influences are applied
    input_status_matrix: np.ndarray  # [n_scenarios x n_inputs; bool]
    unique_scenario_idx: np.ndarray  # [n_scenarios; int] Mapping from input scenario to node scenario idx

    # NODE STATUSES
    n_unique_scenarios: int  # Nr of unique node status scenarios
    status_matrix: np.ndarray  # [n_unique_scenarios x n_nodes; Diag] Node status scenarios


@dataclass
class SelectionChoiceScenarios:
    """
    Dataclass representing status scenarios specific to a selection choice.
    """

    # CHOICE SCENARIO INFO
    choice_nodes: List[SelectionChoiceNode]  # Choice nodes represented by this object
    choice_idx: np.ndarray  # [n_nodes; int] Choice indices (in selection_choice_nodes list)
    nodes_idx: np.ndarray  # [n_nodes; int] Choice node indices (in the influence matrix)
    n_opts: np.ndarray  # [n_nodes; int] Nr of options for each choice
    opt_node_idx: List[np.ndarray]  # [n_nodes -> n_opts; int] Node idx of each option node (in influence matrix)
    unshared_opt_node_idx: List[np.ndarray]  # [n_nodes -> n_opts; int] Idem. but assuming no option nodes are shared

    # INPUT INFLUENCES
    input_node_idx: np.ndarray  # [n_inputs; int] Input option node indices (in the influence matrix)
    input_choice_idx: np.ndarray  # [n_inputs; int] Input choice indices (in selection_choice_nodes list)
    input_opt_idx: np.ndarray  # [n_inputs; int] Input option node indices (wrt choice)
    # Input status matrices, True represents that the input node is selected in the graph, so its influences are applied
    input_status_matrix: np.ndarray  # [n_scenarios x n_inputs; bool]
    unique_scenario_idx: np.ndarray  # [n_scenarios; int] Mapping from input scenario to choice activation scenario

    # OPTION SELECTION INFLUENCE
    n_combinations: np.ndarray  # [n_scenarios; int] Nr of option combinations for each input scenario
    n_total: int  # Total nr of combinations for all input scenarios (sum of n_combinations)
    n_combinations_unique: np.ndarray  # [n_unique_scenarios; int] Nr of option combs for choice activation scenarios
    # Status scenarios before choice selection, for each choice activation scenario, for each choice node, the initial
    # status matrix for 1+n_opts (choice node + option nodes); specifying choice activeness and option availability
    status_matrix: List[List[np.ndarray]]  # [n_unique_scenarios -> n_nodes -> 1+n_opts; Diag]
    # Option-index combinations, for each choice activation scenario, a matrix of n_combinations x n_nodes; specifying
    # in each row a possible option-selection combination, the values are selected option-indices wrt the choice
    opt_idx_combinations: List[np.ndarray]  # [n_unique_scenarios -> n_combs x n_nodes; int]
    # Idem. but values are selected node indices in the influence matrix
    node_idx_combinations: List[np.ndarray]  # [n_unique_scenarios -> n_combs x n_nodes; int]
    # Applied status combinations, representing node status after applying the option-index combination, for each choice
    # activation scenario, for each choice node
    applied_status_combinations: List[List[np.ndarray]]  # [n_unique_scenarios -> n_nodes -> n_combs x 1+n_opts; int]


class InfluenceMatrix:
    """
    Creates an influence matrix, specifying which option-decision node influence the existence of other nodes.
    The matrix is structured with the following elements on the diagonal:
    - For each option-decision (in order of descending importance)
      - The decision node
      - The option nodes (in order)
    - All other nodes
    """

    def __init__(self, adsg: 'DSGType', remove_duplicate_nodes=False):
        if adsg.derivation_start_nodes is None or not adsg.feasible:
            raise ValueError('Provide a feasible graph with external functions!')
        self._remove_duplicate_nodes = remove_duplicate_nodes
        self._adsg = adsg

    @property
    def adsg(self) -> 'DSGType':
        return self._adsg

    @cached_property
    def permanent_nodes(self) -> Set[DSGNode]:
        adsg = self.adsg
        start_nodes = adsg.derivation_start_permanent_nodes
        permanent_nodes, _ = traverse_until_choice_nodes(adsg.graph, start_nodes)
        return permanent_nodes

    @cached_property
    def permanent_nodes_incl_choice_nodes(self) -> Set[DSGNode]:
        adsg = self.adsg
        start_nodes = adsg.derivation_start_permanent_nodes
        permanent_nodes, init_decision_nodes = traverse_until_choice_nodes(adsg.graph, start_nodes)
        return permanent_nodes | init_decision_nodes

    @cached_property
    def choice_nodes(self) -> List[ChoiceNode]:
        adsg = self.adsg
        graph = adsg.graph

        decision_nodes = []
        start_nodes = adsg.derivation_start_nodes
        traversed = set()
        while True:
            # From the current starting nodes, get the next decision nodes
            non_dec_nodes, next_decision_nodes = traverse_until_choice_nodes(graph, start_nodes, traversed=traversed)
            if len(next_decision_nodes) == 0:
                break

            # Order the decision nodes
            decision_nodes += adsg.ordered_choice_nodes(next_decision_nodes)

            # Move to next starting nodes
            traversed |= set(non_dec_nodes) | set(next_decision_nodes)
            start_nodes = next_decision_nodes

        return decision_nodes

    @cached_property
    def selection_choice_nodes(self) -> List[SelectionChoiceNode]:
        return [choice_node for choice_node in self.choice_nodes if isinstance(choice_node, SelectionChoiceNode)]

    @cached_property
    def selection_choice_option_nodes(self) -> Dict[SelectionChoiceNode, List[DSGNode]]:
        return {choice_node: self.adsg.get_option_nodes(choice_node) for choice_node in self.selection_choice_nodes}

    @cached_property
    def _sel_choice_influence(self) \
            -> Dict[SelectionChoiceNode, Dict[DSGNode, Tuple[Set[DSGNode], Set[DSGNode], bool]]]:

        cache = {}

        def _get_confirmed_removed_nodes(sel_choice_node: SelectionChoiceNode, option_node: DSGNode):
            # Confirmed nodes: nodes directly confirmed by this option node
            #  --> (i.e. nodes that will exist if this option node is chosen)
            # Removed nodes: nodes directly removed by this option node
            try:
                confirmed_edges, removed_nodes =\
                    graph.get_confirmed_edges_selection_choice(sel_choice_node, option_node, include_choice=True,
                                                               included_apply_edges=False, cache=cache)
                confirmed_nodes = {node for edge in confirmed_edges for node in edge[:2]}
                confirmed_nodes.add(option_node)
                is_feasible = True

            except IncompatibilityError:
                confirmed_nodes, removed_nodes, is_feasible = set(), set(), False

            return confirmed_nodes, removed_nodes, is_feasible

        graph = self.adsg
        return {node: {option: _get_confirmed_removed_nodes(node, option) for option in options}
                for node, options in self.selection_choice_option_nodes.items()}

    @cached_property
    def other_nodes(self) -> List[DSGNode]:
        decision_nodes = set(self.selection_choice_nodes) | \
                         {node for opt_nodes in self.selection_choice_option_nodes.values() for node in opt_nodes}
        return [node for node in self.adsg.graph.nodes if node not in decision_nodes]

    @cached_property
    def matrix_diagonal_nodes(self) -> List[DSGNode]:
        diagonal_nodes = []

        node_idx_map = {}
        sel_choice_opt_nodes = self.selection_choice_option_nodes
        for sel_choice_node in self.selection_choice_nodes:
            node_idx_map[sel_choice_node] = len(diagonal_nodes)
            diagonal_nodes.append(sel_choice_node)
            for opt_node in sel_choice_opt_nodes[sel_choice_node]:
                node_idx_map[opt_node] = len(diagonal_nodes)
                diagonal_nodes.append(opt_node)

        if self._remove_duplicate_nodes:
            keep_idx = sorted(node_idx_map.values())
            diagonal_nodes = [node for i, node in enumerate(diagonal_nodes) if i in keep_idx]

        diagonal_nodes += self.other_nodes
        return diagonal_nodes

    @cached_property
    def matrix_diagonal_nodes_idx(self) -> Dict[DSGNode, int]:
        return {node: i for i, node in enumerate(self.matrix_diagonal_nodes)}

    @cached_property
    def choice_idx(self) -> np.ndarray:
        idx_map = self.matrix_diagonal_nodes_idx
        return np.array([idx_map[sel_choice_node] for sel_choice_node in self.choice_nodes])

    @cached_property
    def influence_matrix(self) -> np.ndarray:
        """The matrix of n_nodes x n_nodes specifying initial node status and node influence"""
        return self._create_influence_matrix()

    @cached_property
    def influence_matrix_no_choice_constraints(self) -> np.ndarray:
        """The matrix of n_nodes x n_nodes specifying initial node status and node influence"""
        return self._create_influence_matrix(apply_choice_constraints=False)

    @cached_property
    def init_status_array(self) -> np.ndarray:
        """Get the initial status array (before any decisions have been taken)"""
        matrix = self.influence_matrix
        i_diag = np.arange(matrix.shape[0])
        return matrix[i_diag, i_diag]

    def get_next_choice_nodes(self, status_array: np.ndarray) -> List[ChoiceNode]:
        if len(self.choice_idx) == 0:
            return []
        active_choice_node_idx = np.where(status_array[self.choice_idx] == Diag.CONFIRMED.value)[0]
        choice_nodes = self.choice_nodes
        return self.adsg.ordered_choice_nodes([choice_nodes[idx] for idx in active_choice_node_idx])

    def apply_selection_choice(self, status_array: np.ndarray, choice_node: SelectionChoiceNode,
                               option_node: Optional[DSGNode]) -> np.ndarray:

        status_array = status_array.copy()
        i_choice_apply = self.matrix_diagonal_nodes_idx[choice_node]
        i_opt_apply = self.matrix_diagonal_nodes_idx[option_node] if option_node is not None else None

        # Mark which nodes are already confirmed
        not_confirmed_mask = status_array == Diag.INITIAL.value

        # Apply statuses for confirmation and removal influences
        if i_opt_apply is not None:
            matrix = self.influence_matrix
            confirmation_influence_map = matrix[i_opt_apply, :] == OffDiag.CONFIRMATION.value
            status_array[confirmation_influence_map & not_confirmed_mask] = Diag.CONFIRMED.value

            removal_influence_map = matrix[i_opt_apply, :] == OffDiag.REMOVAL.value
            status_array[removal_influence_map & not_confirmed_mask] = Diag.REMOVED.value

        # Mark choice as made
        status_array[i_choice_apply] = Diag.CHOICE_MADE.value
        if i_opt_apply is not None:
            status_array[i_opt_apply] = Diag.CONFIRMED.value

        return status_array

    def apply_choice(self, status_array: np.ndarray, decision_node: ChoiceNode, copy=True) -> np.ndarray:
        if copy:
            status_array = status_array.copy()

        i_choice_apply = self.matrix_diagonal_nodes_idx[decision_node]
        status_array[i_choice_apply] = Diag.CHOICE_MADE.value
        return status_array

    def _create_influence_matrix(self, apply_choice_constraints=True) -> np.ndarray:
        # Initialize matrix
        diag_nodes = self.matrix_diagonal_nodes
        node_idx = self.matrix_diagonal_nodes_idx
        matrix = np.zeros((len(diag_nodes), len(diag_nodes)), dtype=int)

        # Mark permanent nodes and initial decisions
        permanent_nodes = self.permanent_nodes_incl_choice_nodes
        for node in permanent_nodes:
            i_node = node_idx[node]
            matrix[i_node, i_node] = Diag.CONFIRMED.value

        # Loop over choices
        opt_node_map = self.selection_choice_option_nodes
        influence_map = self._sel_choice_influence
        for choice_node in self.selection_choice_nodes:

            # Determine hierarchical influence of all option nodes
            for opt_node in opt_node_map[choice_node]:
                i_opt = node_idx[opt_node]

                opt_confirmed_nodes, opt_removed_nodes, opt_is_feasible = influence_map[choice_node][opt_node]

                # Set feasibility flag
                if not opt_is_feasible:
                    matrix[i_opt, i_opt] = Diag.INFEASIBLE_OPTION.value
                    continue

                # Do not confirm permanent nodes
                opt_confirmed_nodes -= permanent_nodes

                # Set influence flags
                for nodes, flag in [(opt_confirmed_nodes, OffDiag.CONFIRMATION.value),
                                    (opt_removed_nodes, OffDiag.REMOVAL.value)]:
                    for node in nodes:
                        # Only if it is not referring to the option node
                        j_removed_node = node_idx[node]
                        if j_removed_node == i_opt:
                            continue
                        matrix[i_opt, j_removed_node] = flag

        # Apply choice constraints
        if apply_choice_constraints:
            other_nodes = self.other_nodes
            for choice_constraint in self.adsg.get_choice_constraints():
                for i_dec, choice_node in enumerate(choice_constraint.nodes):
                    if not isinstance(choice_node, SelectionChoiceNode):
                        break

                    for i_opt, opt_node in enumerate(choice_constraint.options[i_dec]):
                        if opt_node not in node_idx:
                            continue
                        i_src = node_idx[opt_node]
                        for _, removed_opt_nodes in get_constraint_removed_options(choice_constraint, i_dec, i_opt):
                            for removed_opt_node in removed_opt_nodes:
                                if removed_opt_node not in node_idx or removed_opt_node in other_nodes:
                                    continue
                                i_tgt = node_idx[removed_opt_node]
                                matrix[i_src, i_tgt] = OffDiag.REMOVAL.value

        # Clear feedback influences from non-choice nodes to choice nodes
        i_other = np.array(sorted([node_idx[node] for node in self.other_nodes]), dtype=int)
        i_choice_nodes = np.ones((len(diag_nodes),), dtype=bool)
        i_choice_nodes[i_other] = False
        i_choice_nodes = np.where(i_choice_nodes)[0]
        matrix[np.ix_(i_other, i_choice_nodes)] = OffDiag.NO_INTERACTION.value

        return matrix

    @cached_property
    def base_sel_choice_scenarios(self):
        """
        Get the selection choice scenarios without merging.
        """
        node_idx_map = self.matrix_diagonal_nodes_idx
        opt_node_map = self.selection_choice_option_nodes

        choice_idx_map = {node: i_choice for i_choice, node in enumerate(self.selection_choice_nodes)}
        opt_idx_map = {node_idx_map[opt_node]: (choice_idx_map[choice_node], i_opt)
                       for choice_node, opt_nodes in opt_node_map.items() for i_opt, opt_node in enumerate(opt_nodes)}

        # Loop over each choice node
        choice_scenarios = []
        for i_choice, sel_choice_node in enumerate(self.selection_choice_nodes):
            node_idx = node_idx_map[sel_choice_node]

            opt_nodes = opt_node_map[sel_choice_node]
            opt_nodes_idx = np.array([node_idx_map[opt_node] for opt_node in opt_nodes])
            unshared_opt_nodes_idx = opt_nodes_idx
            unshared_opt_nodes_idx = np.arange(len(unshared_opt_nodes_idx))+node_idx+1

            choice_opt_nodes: List[DSGNode] = [sel_choice_node]
            choice_opt_nodes += opt_nodes

            # Get status scenarios for the choice node and option nodes
            status_scenario = self.get_status_scenarios(choice_opt_nodes)
            input_node_idx = status_scenario.influence_input_node_idx
            if len(input_node_idx) > 0:
                input_choice_idx, input_opt_idx = zip(*[opt_idx_map[node_idx] for node_idx in input_node_idx])
                input_choice_idx = np.array(input_choice_idx)
                input_opt_idx = np.array(input_opt_idx)
            else:
                input_choice_idx = np.array([], dtype=int)
                input_opt_idx = np.array([], dtype=int)
            input_status_matrix = status_scenario.input_status_matrix
            unique_scenario_idx = status_scenario.unique_scenario_idx

            # For each status scenario, determine whether the choice is active
            status_matrix = status_scenario.status_matrix
            choice_is_active = np.array(status_matrix[:, 0] == Diag.CONFIRMED.value)

            # Remove duplicate choice-inactive scenarios
            is_inactive = np.where(~choice_is_active)[0]
            if len(is_inactive) > 1:
                # Ignore removal interactions for choice-inactive scenarios
                status_matrix = status_matrix.copy()
                status_matrix_inactive = status_matrix[~choice_is_active, 1:]
                status_matrix_inactive[status_matrix_inactive == OffDiag.REMOVAL.value] = OffDiag.NO_INTERACTION.value
                status_matrix[~choice_is_active, 1:] = status_matrix_inactive

                # Get unique remaining scenarios
                status_matrix, selected_idx, selected_scenario_idx = \
                    np.unique(status_matrix, axis=0, return_inverse=True, return_index=True)
                choice_is_active = choice_is_active[selected_idx]
                unique_scenario_idx = selected_scenario_idx[unique_scenario_idx]

            # For each status scenario determine which options are available
            opt_is_available = ~((status_matrix[:, 1:] == Diag.REMOVED.value) |
                                 (status_matrix[:, 1:] == Diag.INFEASIBLE_OPTION.value))

            # Create option idx combinations
            opt_idx_combinations = []
            node_idx_combinations = []
            applied_status_combinations = []
            for i_stat, is_active in enumerate(choice_is_active):
                applied_status = status_matrix[[i_stat], :].copy()

                if not is_active:
                    opt_idx_combinations.append(np.array([[X_INACTIVE_VALUE]], dtype=int))
                    node_idx_combinations.append(np.array([[X_INACTIVE_VALUE]], dtype=int))
                    applied_status_combinations.append([applied_status])

                else:
                    opt_avail = np.where(opt_is_available[i_stat, :])[0]
                    opt_idx_combinations.append(np.array([opt_avail]).T)
                    node_idx_combinations.append(np.array([opt_nodes_idx[opt_avail]]).T)

                    # Mark choice as made
                    applied_status[0, 0] = Diag.CHOICE_MADE.value

                    # Mark chosen option nodes as selected (confirmed)
                    applied_status = np.repeat(applied_status, repeats=len(opt_avail), axis=0)
                    for i_app, applied_idx in enumerate(opt_avail):
                        applied_status[i_app, applied_idx+1] = Diag.CONFIRMED.value
                    applied_status_combinations.append([applied_status])

            n_combinations_unique = np.array([comb.shape[0] for comb in opt_idx_combinations])
            n_combinations = n_combinations_unique[unique_scenario_idx]

            choice_scenarios.append(SelectionChoiceScenarios(
                choice_nodes=[sel_choice_node],
                choice_idx=np.array([i_choice]),
                nodes_idx=np.array([node_idx]),
                n_opts=np.array([len(opt_nodes)]),
                opt_node_idx=[opt_nodes_idx],
                unshared_opt_node_idx=[unshared_opt_nodes_idx],
                input_node_idx=input_node_idx,
                input_choice_idx=input_choice_idx,
                input_opt_idx=input_opt_idx,
                input_status_matrix=input_status_matrix,
                unique_scenario_idx=unique_scenario_idx,
                n_combinations=n_combinations,
                n_total=int(np.sum(n_combinations)),
                status_matrix=[[row] for row in status_matrix],
                n_combinations_unique=n_combinations_unique,
                opt_idx_combinations=opt_idx_combinations,
                node_idx_combinations=node_idx_combinations,
                applied_status_combinations=applied_status_combinations,
            ))
        return choice_scenarios

    def get_status_scenarios(self, nodes: List[DSGNode]) -> StatusScenarios:
        """
        Get independent status scenarios for given nodes. A status scenario is a set of status for the given nodes given
        some input influence (confirmation, removal) statuses.
        """
        node_idx_map = self.matrix_diagonal_nodes_idx
        nodes_idx = np.array([node_idx_map[node] for node in nodes])
        influence_matrix = self.influence_matrix_no_choice_constraints

        no_interaction = OffDiag.NO_INTERACTION.value
        confirms_interaction = OffDiag.CONFIRMATION.value
        removes_interaction = OffDiag.REMOVAL.value
        initial_status = Diag.INITIAL.value
        confirmed_status = Diag.CONFIRMED.value
        removed_status = Diag.REMOVED.value

        # Get input influences
        matrix_inputs = influence_matrix[:, nodes_idx].copy()
        matrix_inputs[nodes_idx, np.arange(len(nodes_idx))] = no_interaction
        influence_input_node_idx = np.where(np.any(matrix_inputs != no_interaction, axis=1))[0]
        is_internal = np.zeros((matrix_inputs.shape[0],), dtype=bool)
        is_internal[nodes_idx] = True
        permanent_nodes = np.diag(influence_matrix) == Diag.CONFIRMED.value

        # If there are no influences, there is one scenario: the starting scenario
        starting_status = np.diag(influence_matrix)[nodes_idx]
        if len(influence_input_node_idx) == 0:
            status_matrix = np.array([starting_status])
            input_status_matrix = np.zeros((1, 0), dtype=bool)

        else:
            # Get the Cartesian product of all input influence existence scenarios
            status_matrix = []
            input_status_matrix = []
            for input_exists in itertools.product(*[[True] if permanent_nodes[node_idx] else [False, True]
                                                    for node_idx in influence_input_node_idx]):

                input_exists_arr = np.array(input_exists)
                input_status_matrix.append(input_exists_arr)

                # Apply the input influences to modify the node statuses
                node_status = starting_status.copy()
                selected_inputs = influence_input_node_idx[input_exists_arr]
                applied_influences = matrix_inputs[selected_inputs, :]
                applied_is_internal = is_internal[selected_inputs]
                if applied_influences.shape[0] > 0:
                    is_initial_mask = starting_status == initial_status

                    # First apply internal confirming influences (relevant if selecting one option node also confirms
                    # another option node, e.g. for component instantiation)
                    is_confirmed = is_initial_mask & \
                        np.any(applied_influences[applied_is_internal] == confirms_interaction, axis=0)
                    node_status[is_confirmed] = confirmed_status

                    # Apply external influences
                    is_removed = is_initial_mask & ~is_confirmed & \
                        np.any(applied_influences == removes_interaction, axis=0)
                    node_status[is_removed] = removed_status

                    is_confirmed = is_initial_mask & ~is_removed & \
                        np.any(applied_influences == confirms_interaction, axis=0)
                    node_status[is_confirmed] = confirmed_status

                status_matrix.append(node_status)

            status_matrix = np.array(status_matrix)
            input_status_matrix = np.array(input_status_matrix)

        # Get unique scenarios
        status_matrix, unique_scenario_idx = np.unique(status_matrix, axis=0, return_inverse=True)

        return StatusScenarios(
            nodes=nodes,
            nodes_idx=nodes_idx,
            n_scenarios=len(unique_scenario_idx),
            n_unique_scenarios=status_matrix.shape[0],
            influence_input_node_idx=influence_input_node_idx,
            input_status_matrix=input_status_matrix,
            unique_scenario_idx=unique_scenario_idx,
            status_matrix=status_matrix,
        )



import numpy as np
from typing import *
from adsg_core.optimization.assign_enc.encoding import *

__all__ = ['ConstraintViolationImputer']


class ConstraintViolationImputer(EagerImputer):
    """Does not do imputation, but sends a matrix with -1 as values to indicate an invalid design point.
    The optimization problem should represent this as a violated constraint."""

    def impute(self, vector: DesignVector, existence: NodeExistence, matrix_mask: MatrixSelectMask) -> Tuple[DesignVector, np.ndarray]:
        matrix = self._get_matrix(existence)
        if matrix.shape[0] == 0:
            return vector, np.zeros((0, 0), dtype=int)
        invalid_matrix = matrix[0, :, :]*0-1
        return vector, invalid_matrix

    def __repr__(self):
        return f'{self.__class__.__name__}()'

    def __str__(self):
        return f'Constr Vio Imp'



"""
MIT License

Copyright: (c) 2024, Deutsches Zentrum fuer Luft- und Raumfahrt e.V.
Contact: jasper.bussemaker@dlr.de

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""
import pickle
import hashlib
import functools


def cached_function(func):
    """
    Function decorator for caching function calls using the class dict, so it can be serialized (pickled).

    Inspired by:
    - https://stackoverflow.com/a/15587418
    - cached_property
    """
    name = func.__name__

    def wrapper(obj, *args, **kwargs):
        if obj is None:
            raise RuntimeError('cached_serializable should be applied to a regular class function,'
                               'not a @staticmethod or @classmethod!')
        cache = obj.__dict__

        # Python-session persistent hashing, because the hash() function is randomized between Python instances
        # This is slower than normal hashing, so that's why we also apply lru_cache below
        cache_args = b'||'.join((b'__'.join(pickle.dumps(v) for v in args),
                                 b'__'.join(str(k).encode('utf-8')+b'='+pickle.dumps(v) for k, v in kwargs.items())))
        cache_key = '_cache_'+name+'_'+hashlib.md5(cache_args).hexdigest()[:8]

        if cache_key in cache:
            return cache[cache_key]
        cache[cache_key] = value = func(obj, *args, **kwargs)
        return value

    wrapper.__name__ = name
    return functools.lru_cache()(wrapper)


def clear_func_cache(obj, func=None):
    key_start = '_cache_'
    if func is not None:
        func_name = func if isinstance(func, str) else func.__name__
        key_start += func_name+'_'

    # Clear persistent cache
    cleared_funcs = set()
    for key in list(obj.__dict__.keys()):
        if key.startswith(key_start):
            cleared_funcs.add('_'.join(key.split('_')[2:-1]))
            del obj.__dict__[key]

    # Clear lru_cache
    for cleared_func_name in cleared_funcs:
        func = getattr(obj, cleared_func_name)
        if hasattr(func, 'cache_clear'):
            func.cache_clear()



import os
import pickle
import logging
import numpy as np
import pandas as pd
from typing import *
from adsg_core.optimization.assign_enc.matrix import *
from adsg_core.optimization.assign_enc.encoding import *
from adsg_core.optimization.assign_enc.time_limiter import *
from adsg_core.optimization.assign_enc.lazy_encoding import *
from adsg_core.optimization.assign_enc.encoder_registry import *
from adsg_core.optimization.assign_enc.patterns.encoder import *
from adsg_core.optimization.assign_enc.assignment_manager import *
from adsg_core.optimization.assign_enc.cache import get_cache_path

__all__ = ['EncoderSelector']

log = logging.getLogger('adsg_core.selector')


class EncoderSelector:
    """Algorithm for automatically selecting the best encoder for a given assignment problem."""

    _global_disable_cache = False  # For testing/experiments
    _exclude_pattern_encoders = False
    _print_stats = False
    _numba_initialized = False

    encoding_timeout = .25  # sec
    n_mat_max_eager = 1e3
    imputation_ratio_limits = [10, 40, 100]

    limit_dist_corr_time = True
    min_distance_correlation = .7

    def __init__(self, settings: MatrixGenSettings):
        self.settings: MatrixGenSettings = settings

        self.lazy_imputer = DEFAULT_LAZY_IMPUTER
        self.eager_imputer = DEFAULT_EAGER_IMPUTER
        self._last_selection_stage = None

    def reset_cache(self):
        cache_path = self._cache_path(f'{self._get_cache_key()}.pkl')
        if os.path.exists(cache_path):
            os.remove(cache_path)

    def get_best_assignment_manager(self, cache=True, limit_time=True) -> AssignmentManagerBase:
        cache_path = self._cache_path(f'{self._get_cache_key()}.pkl')
        if not self._global_disable_cache and cache and os.path.exists(cache_path):
            with open(cache_path, 'rb') as fp:
                return pickle.load(fp)
        self.initialize_numba()

        enc_timeout, n_mme, limit_dc_time = self.encoding_timeout, self.n_mat_max_eager, self.limit_dist_corr_time
        if not limit_time:
            self.encoding_timeout = 10
            self.n_mat_max_eager = 1e5
            # self.limit_dist_corr_time = False

        assignment_manager = self._get_best_assignment_manager()

        if not limit_time:
            self.encoding_timeout, self.n_mat_max_eager, self.limit_dist_corr_time = enc_timeout, n_mme, limit_dc_time

        os.makedirs(os.path.dirname(cache_path), exist_ok=True)
        with open(cache_path, 'wb') as fp:
            pickle.dump(assignment_manager, fp)
        return assignment_manager

    def initialize_numba(self):
        """Run a simple encoding problem so that numba functions are JIT compiled and compilation time doesn't influence
        selection timeouts"""
        if self.__class__._numba_initialized:
            return
        self.__class__._numba_initialized = True

        enc_timeout, settings = self.encoding_timeout, self.settings

        self.encoding_timeout = 10
        self.settings = MatrixGenSettings(
            src=[Node([0, 1]) for _ in range(2)], tgt=[Node([0, 1]) for _ in range(2)],
        )
        log.debug('Initializing numba...')
        try:
            self._exclude_pattern_encoders = True
            self._get_best_assignment_manager()
        finally:
            self.encoding_timeout, self.settings = enc_timeout, settings
            self._exclude_pattern_encoders = False

    def _get_best_assignment_manager(self) -> AssignmentManager:
        log.debug('Counting matrices...')
        self._provision_agg_matrix_cache()
        n_mat, n_exist = self._get_n_mat()
        log.debug(f'{n_mat} matrices ({n_exist} existence scheme{"s" if n_exist != 1 else ""})')

        dist_corr_lazy_limit = self.imputation_ratio_limits[0]
        dist_corr_eager_limit = self.imputation_ratio_limits[-1]
        high_imp_ratio_detect_limit = self.imputation_ratio_limits[-1]

        def _instantiate_manager(encoder):
            cls = LazyAssignmentManager if isinstance(encoder, LazyEncoder) else AssignmentManager
            return cls(self.settings, encoder)

        def _create_managers(encoders, imputer_factory, dist_corr_limit, limit_dist_corr_time=None):
            assignment_mgr = []
            scoring = {'n_des_pts': [], 'imp_ratio': [], 'inf_idx': [], 'dist_corr': []}
            with Encoder.with_early_detect_high_imp_ratio(high_imp_ratio=high_imp_ratio_detect_limit):
                for encoder_factory in encoders:
                    # Create encoder
                    encoder = encoder_factory(imputer_factory())

                    # Create assignment manager
                    log.debug(f'Encoding {encoder!s}')
                    try:
                        assignment_manager = run_timeout(self.encoding_timeout, _instantiate_manager, encoder)

                    except InvalidPatternEncoder:
                        continue

                    except DetectedHighImpRatio:
                        log.debug('Excessively high imputation ratio detected!')
                        continue

                    except (TimeoutError, MemoryError):
                        log.debug('Encoding timeout!')
                        continue

                    # Get metrics
                    n_design_points = assignment_manager.encoder.get_n_design_points()
                    imputation_ratio = self._get_imp_ratio(
                        n_design_points, n_mat, n_exist, assignment_manager=assignment_manager)
                    information_index = assignment_manager.encoder.get_information_index()

                    distance_correlation = np.nan
                    if imputation_ratio <= dist_corr_limit:
                        if limit_dist_corr_time:
                            limit_dist_corr_time = self.limit_dist_corr_time
                        if limit_dist_corr_time:
                            try:
                                distance_correlation = run_timeout(
                                    self.encoding_timeout, self._get_dist_corr, assignment_manager)
                            except (TimeoutError, MemoryError):
                                pass
                        else:
                            distance_correlation = self._get_dist_corr(assignment_manager)

                    assignment_mgr.append(assignment_manager)
                    scoring['n_des_pts'].append(n_design_points)
                    scoring['imp_ratio'].append(imputation_ratio)
                    scoring['inf_idx'].append(information_index)
                    scoring['dist_corr'].append(distance_correlation)

                    # Stop if we found a perfect encoder
                    if imputation_ratio == 1. and distance_correlation >= .95:
                        break

            # Equalize distance correlations
            df = pd.DataFrame(data=scoring)
            unique_values, unique_indices_list = np.unique(df[['imp_ratio', 'inf_idx']].values, axis=0, return_inverse=True)
            dist_corr_values = df.dist_corr.values
            for i_unique in range(len(unique_values)):
                unique_indices, = np.where(unique_indices_list == i_unique)
                if len(unique_indices) > 1:
                    values = df.dist_corr.values[unique_indices]
                    values = values[~np.isnan(values)]
                    if len(values) > 0:
                        dist_corr_values[unique_indices] = np.mean(values)
            df.dist_corr = pd.Series(index=df.dist_corr.index, data=np.round(dist_corr_values/.01)*.01)

            return df, assignment_mgr

        # Special case if there are no possible connections
        if n_mat == 0:
            return _instantiate_manager(DEFAULT_EAGER_ENCODER())

        def _print_stats(i_select):
            if not self._print_stats:
                return
            df_score['name'] = [str(am.encoder) for am in assignment_managers]
            df_score['selection_stage'] = [self._last_selection_stage if i_am == i_select else ''
                                           for i_am in range(len(df_score))]
            pd.set_option('display.max_columns', None)
            pd.set_option('display.max_rows', None)
            pd.set_option('display.expand_frame_repr', False)
            print(df_score)

        # Try pattern encoders
        df_score, assignment_managers = None, []
        if not self._exclude_pattern_encoders:
            df_score, assignment_managers = _create_managers(
                PATTERN_ENCODERS, self.lazy_imputer, dist_corr_lazy_limit, limit_dist_corr_time=False)
            i_best = self._get_best(df_score, knows_n_mat=n_mat is not None, n_priority=5)
            if i_best is not None:
                self._last_selection_stage = '0_pattern'
                _print_stats(i_best)
                return assignment_managers[i_best]

        # If there are not too many matrices, encode all encoders except enumeration-based
        initially_all = False
        if n_mat is not None and n_mat <= self.n_mat_max_eager:
            df_eager, eager_assignment_managers = \
                _create_managers(EAGER_ENCODERS, self.eager_imputer, dist_corr_eager_limit)
            df_score = pd.concat([df_score, df_eager], ignore_index=True) if df_score is not None else df_eager

            df_lazy, lazy_assignment_mgr = _create_managers(LAZY_ENCODERS, self.lazy_imputer, dist_corr_lazy_limit)
            df_score = pd.concat([df_score, df_lazy], ignore_index=True)
            assignment_managers += eager_assignment_managers+lazy_assignment_mgr
            initially_all = True

        # Otherwise initially only try lazy encoders
        else:
            df_lazy, lazy_assignment_managers = _create_managers(LAZY_ENCODERS, self.lazy_imputer, dist_corr_lazy_limit)
            df_score = pd.concat([df_score, df_lazy], ignore_index=True) if df_score is not None else df_lazy
            assignment_managers += lazy_assignment_managers

        i_best = self._get_best(df_score, knows_n_mat=n_mat is not None, n_priority=4)
        if i_best is not None:
            self._last_selection_stage = '1_init_all' if initially_all else '1_init_lazy'
            _print_stats(i_best)
            return assignment_managers[i_best]

        # Select based on information index (higher information indices correlate with higher distance correlation for
        # non-enumeration-based encoders)
        i_best = self._get_best(df_score, knows_n_mat=n_mat is not None, n_priority=4, by_inf_idx=True)
        if i_best is not None:
            self._last_selection_stage = '2_init_inf_idx'
            _print_stats(i_best)
            return assignment_managers[i_best]

        # If none found, add the other type and remaining priority areas
        if not initially_all:
            df_other, other_assignment_mgr = _create_managers(EAGER_ENCODERS, self.eager_imputer, dist_corr_eager_limit)
            df_score = pd.concat([df_other, df_score], ignore_index=True)
            assignment_managers = other_assignment_mgr+assignment_managers

            i_best = self._get_best(df_score, knows_n_mat=n_mat is not None)
            if i_best is not None:
                self._last_selection_stage = '3_all'
                _print_stats(i_best)
                return assignment_managers[i_best]

        # Also include enumeration-based encoders
        df_enum, enum_assignment_mgr = _create_managers(EAGER_ENUM_ENCODERS, self.lazy_imputer, dist_corr_eager_limit)
        if np.all(np.isnan(df_score.dist_corr)):
            df_enum.dist_corr = np.nan
        df_score = pd.concat([df_score, df_enum], ignore_index=True)
        assignment_managers = assignment_managers+enum_assignment_mgr

        i_best = self._get_best(df_score, knows_n_mat=n_mat is not None)
        if i_best is not None:
            self._last_selection_stage = '4_all_enum'
            _print_stats(i_best)
            return assignment_managers[i_best]

        # Select based on information index
        i_best = self._get_best(df_score, knows_n_mat=n_mat is not None, by_inf_idx=True)
        if i_best is None:
            raise RuntimeError(f'Cannot find best encoder, try increasing timeout')
        self._last_selection_stage = '4_all_inf_idx'
        _print_stats(i_best)
        return assignment_managers[i_best]

    def _get_imp_ratio(self, n_design_points: int, n_mat: int = None, n_exist: int = None,
                       assignment_manager: AssignmentManagerBase = None) -> float:
        if assignment_manager is not None:
            return assignment_manager.encoder.get_imputation_ratio(per_existence=True)
        return ((n_design_points*n_exist)/n_mat) if n_mat is not None else n_design_points

    def _get_dist_corr(self, assignment_manager: AssignmentManagerBase) -> float:
        return assignment_manager.encoder.get_distance_correlation(minimum=True)

    def reset_agg_matrix_cache(self):
        self._get_matrix_gen().reset_agg_matrix_cache()

    def _provision_agg_matrix_cache(self):
        log.debug('Generating aggregate matrix for eager encoders...')
        self._get_matrix_gen().get_agg_matrix(cache=True)

    def _get_best(self, df_scores: pd.DataFrame, knows_n_mat, n_priority: int = None,
                  by_inf_idx=False) -> Optional[int]:
        if len(df_scores) == 0:
            return
        df_scores = df_scores.copy()
        df_scores['idx'] = df_scores.index

        corr_col = 'inf_idx' if by_inf_idx else 'dist_corr'

        # If we don't accurately know the total amount of matrices, normalize to the lowest imputation ratio
        if not knows_n_mat:
            df_scores.imp_ratio /= df_scores.imp_ratio.min()

        def _return_best_within_priority_area(df_):
            if by_inf_idx:
                # Get rows with the minimum imputation ratio
                min_imp_ratio = np.min(df_.imp_ratio)
                min_imp_ratio_mask = df_.imp_ratio == min_imp_ratio

                # Within these rows, get the row with the maximum information index
                i_max_inf_idx = np.argmax(df_.inf_idx[min_imp_ratio_mask])
                min_imp_idx = df_.idx.values[min_imp_ratio_mask]
                return int(min_imp_idx[i_max_inf_idx])

            # Get rows with maximum distance correlation
            max_dist_corr = np.nanmax(df_.dist_corr)
            if np.isnan(max_dist_corr):
                return
            max_dist_corr_mask = df_.dist_corr == max_dist_corr

            # Within these rows, get the row with the minimum imputation ratio
            i_min_imp_rat = np.nanargmin(df_.imp_ratio[max_dist_corr_mask])
            min_imp_idx = df_.idx.values[max_dist_corr_mask]
            return int(min_imp_idx[i_min_imp_rat])

        # Check for points with low imputation ratio first
        df_imp_bands = [df_scores[df_scores.imp_ratio == 1]]
        df_imp_bands += [df_scores[df_scores.imp_ratio <= ir_limit] for ir_limit in self.imputation_ratio_limits]
        df_imp_bands += [df_scores]

        if by_inf_idx:
            df_priority = [
                df_imp_bands[0][df_imp_bands[0][corr_col] >= self.min_distance_correlation],
                df_imp_bands[1][df_imp_bands[1][corr_col] >= self.min_distance_correlation],
                df_imp_bands[0][df_imp_bands[0][corr_col] >= .5 * self.min_distance_correlation],
                df_imp_bands[1][df_imp_bands[1][corr_col] >= .5 * self.min_distance_correlation],
                df_imp_bands[0][df_imp_bands[0][corr_col] > 0],
                df_imp_bands[1][df_imp_bands[1][corr_col] > 0],
                df_imp_bands[0],
                df_imp_bands[1],
            ]
        else:
            df_priority = [
                df_imp_bands[0][df_imp_bands[0][corr_col] >= self.min_distance_correlation],
                df_imp_bands[1][df_imp_bands[1][corr_col] >= self.min_distance_correlation],
                df_imp_bands[0][df_imp_bands[0][corr_col] >= .5 * self.min_distance_correlation],
                df_imp_bands[1][df_imp_bands[1][corr_col] >= .5 * self.min_distance_correlation],
                # df_imp_bands[1][(df_imp_bands[1][corr_col] >= 0) & (df_imp_bands[1].inf_idx > 0)],
                df_imp_bands[1][df_imp_bands[1][corr_col] >= 0],
            ]
        for i in range(2, len(df_imp_bands)):
            df_priority += [
                df_imp_bands[i][df_imp_bands[i][corr_col] >= self.min_distance_correlation],
                df_imp_bands[i][df_imp_bands[i][corr_col] >= .5 * self.min_distance_correlation],
                df_imp_bands[i][df_imp_bands[i][corr_col] > 0],
                df_imp_bands[i],
            ]

        for i, df in enumerate(df_priority):
            if n_priority is not None and i >= n_priority:
                return
            if len(df) > 0:
                return _return_best_within_priority_area(df)
        raise RuntimeError('No encoders available!')

    def _get_n_mat(self) -> Tuple[Optional[int], Optional[int]]:
        try:
            matrix_gen = self._get_matrix_gen()
            n_mat_total = run_timeout(
                self.encoding_timeout, lambda: matrix_gen.count_all_matrices(max_by_existence=False))
            n_existence = len(list(matrix_gen.iter_existence()))
            return n_mat_total, n_existence
        except TimeoutError:
            return None, None

    def _get_matrix_gen(self) -> AggregateAssignmentMatrixGenerator:
        return AggregateAssignmentMatrixGenerator(self.settings)

    def _get_cache_key(self):
        return self.settings.get_cache_key()

    def _cache_path(self, sub_path=None):
        sel_cache_folder = 'encoder_cache'
        sub_path = os.path.join(sel_cache_folder, sub_path) if sub_path is not None else sel_cache_folder
        return get_cache_path(sub_path=sub_path)



"""
MIT License

Copyright: (c) 2024, Deutsches Zentrum fuer Luft- und Raumfahrt e.V.
Contact: jasper.bussemaker@dlr.de

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""
import numpy as np
from typing import *
from adsg_core.graph.adsg_nodes import *
from adsg_core.graph.adsg import DSGType
from adsg_core.graph.choice_constraints import *
from adsg_core.optimization.hierarchy.base import *
from adsg_core.optimization.hierarchy.registry import SelChoiceEncoderType

from adsg_core.optimization.sel_choice_enc.encoder import SelectionChoiceEncoder
from adsg_core.optimization.sel_choice_enc.brute_force import BruteForceSelectionChoiceEncoder
from adsg_core.optimization.sel_choice_enc.util import assert_behavior

__all__ = ['SelChoiceEncHierarchyAnalyzer']


class SelChoiceEncHierarchyAnalyzer(HierarchyAnalyzerBase):

    def __init__(self, adsg: DSGType):
        super().__init__(adsg, remove_duplicate_nodes=True)

        node_idx_map = self.influence_matrix.matrix_diagonal_nodes_idx
        self._nodes = sorted(node_idx_map.keys(), key=lambda n: node_idx_map[n])
        self._node_id_map = {node: node_id for node_id, node in enumerate(self._nodes)}
        self._encoder = self._create_sel_choice_encoder()

    def get_encoder_type(self) -> SelChoiceEncoderType:
        return SelChoiceEncoderType.COMPLETE

    def _create_sel_choice_encoder(self) -> SelectionChoiceEncoder:
        influence_matrix = self.influence_matrix.influence_matrix_no_choice_constraints
        choice_constraint_map = self._get_choice_constraint_map()

        node_id_map = self._node_id_map
        node_ids = np.array([node_id_map[node] for node in self.influence_matrix.matrix_diagonal_nodes])

        choice_opt_map = {node_id_map[choice_node]: [node_id_map[opt_node] for opt_node in opt_nodes]
                          for choice_node, opt_nodes in self.influence_matrix.selection_choice_option_nodes.items()}

        return self._instantiate_encoder(influence_matrix, node_ids, choice_opt_map, choice_constraint_map)

    def _get_choice_constraint_map(self) -> Dict[Tuple[int, int], List[Tuple[int, int]]]:
        node_id_map = self._node_id_map
        choice_constraint_map = {}
        for choice_constraint in self.adsg.get_choice_constraints():
            for i_choice, choice_node in enumerate(choice_constraint.nodes):
                if choice_node not in node_id_map:
                    break
                choice_id = node_id_map[choice_node]
                if not isinstance(choice_node, SelectionChoiceNode):
                    break

                for i_opt, opt_node in enumerate(choice_constraint.options[i_choice]):
                    if opt_node not in node_id_map:
                        continue
                    opt_node_id = node_id_map[opt_node]

                    removed_map = []
                    for other_choice_node, removed_opt_nodes in \
                            get_constraint_removed_options(choice_constraint, i_choice, i_opt):

                        if other_choice_node not in node_id_map:
                            continue
                        other_choice_id = node_id_map[other_choice_node]

                        for removed_opt_node in removed_opt_nodes:
                            if removed_opt_node not in node_id_map:
                                continue
                            removed_opt_node_id = node_id_map[removed_opt_node]
                            removed_map.append((other_choice_id, removed_opt_node_id))

                    choice_constraint_map[choice_id, opt_node_id] = removed_map
        return choice_constraint_map

    def _instantiate_encoder(self, influence_matrix, node_ids, choice_opt_map, choice_constraint_map) \
            -> SelectionChoiceEncoder:
        return BruteForceSelectionChoiceEncoder(influence_matrix, node_ids, choice_opt_map, choice_constraint_map)

    def _get_n_opts(self) -> List[int]:
        return self._encoder.design_variable_options

    def _get_selection_choice_is_forced(self) -> np.ndarray:
        return np.array(self._encoder.design_variable_is_forced, dtype=bool)

    def _get_n_combinations(self) -> int:
        return self._encoder.n_valid

    def _get_nodes_existence(self, nodes: List[DSGNode], i_comb: int = None) -> np.ndarray:
        node_ids = self._get_node_ids(nodes)
        return self._encoder.get_node_existence_mask(node_ids, i_dv=i_comb)

    def _get_available_combinations_mask(self, fixed_comb_idx: Dict[int, int]) -> Optional[np.ndarray]:
        if len(fixed_comb_idx) == 0:
            return

        dv_idx = self._encoder.get_design_vector_indices(fixed_comb_idx)
        dv_mask = np.zeros((self.n_combinations,), dtype=bool)
        dv_mask[list(dv_idx)] = True
        return dv_mask

    def _get_comb_idx(self, opt_idx: List[int], include_mask: np.ndarray = None) \
            -> Tuple[Optional[int], Optional[np.ndarray]]:

        i_dv_excluded = None
        if include_mask is not None:
            i_dv_excluded = set(np.where(~include_mask)[0])

        try:
            dv_matched, i_matched = self._encoder.get_design_vector_index(opt_idx, i_dv_excluded=i_dv_excluded)
            return i_matched, np.array(dv_matched)
        except RuntimeError:
            return None, None

    def _get_choice_option_indices(self) -> np.ndarray:
        return self._encoder.all_design_vectors_and_statuses[0]

    def _get_existence_array(self) -> np.ndarray:
        return self._encoder.all_design_vectors_and_statuses[1]

    def _get_node_ids(self, nodes: List[DSGNode]) -> List[int]:
        node_id_map = self._node_id_map
        return [node_id_map[node] for node in nodes]

    def _assert_behavior(self):
        graph = super()._assert_behavior()
        assert_behavior(self._encoder)
        return graph
